# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

import abc
from collections import defaultdict, namedtuple
from itertools import groupby
from operator import attrgetter

from kinder.core.unit.models import Unit, UnitKind
from .constants import EXCLUDED_STATUSES, UNIT_FIELDS

TotalUnit = namedtuple('TotalUnit', 'id, name, count')


class Total(object):
    """Базовый класс итогового подсчета."""
    __metaclass__ = abc.ABCMeta

    def __init__(self, strategy):
        self._strategy = strategy

    @abc.abstractmethod
    def __call__(self, data, period):
        """Точка запуска вычислений показателей по МО."""


class SumTotal(Total):
    """Суммарный подсчет показателей."""

    def __call__(self, data, period):
        """
        Обрабатывает подсчитанные данные, разбитые по учреждениям.

        В каждом учреждении содержится данные о МО к которому он
        принадлежит.

        :param data: tuple(:class:`.types.DOO`)
        """
        # Группирует по аттрибутам название и id МО
        get_key = attrgetter(
            'control_unit_id', 'control_unit')
        grouped = groupby(sorted(data, key=get_key), get_key)
        for (mo_id, mo_name), items in grouped:
            yield TotalUnit(
                id=mo_id,
                name=mo_name,
                count=sum(i.count for i in items))


class CustomTotal(Total):
    """
    Базовый класс сложного итогового подсчета.

    Например в разрезе ДОО должны учитываться все ЖУ ребенка,
    а в разрезе МО нужно учитывать только первое ЖУ из заявки.
    Для реализации нужно будет описать логику в методе
    _calculate, куда приходит информация по МО
    """

    @abc.abstractmethod
    def _calculate(self, mo, period):
        """Подсчет показателя по МО."""

    def __call__(self, data, period):
        mos = Unit.objects.filter(
            id__in=set(d.control_unit_id for d in data))
        for mo in mos:
            yield TotalUnit(
                id=mo.id,
                name=mo.name,
                count=self._calculate(mo, period))


class NumberChildWantChangeTotal(CustomTotal):

    def _calculate(self, mo, period):
        """возвращает только уникальные заявки по МО."""
        mo_units = mo.get_descendants().filter(
            kind_id=UnitKind.DOU,
        ).exclude(
            status__in=EXCLUDED_STATUSES,
            id=mo.id,
        ).values(*UNIT_FIELDS)
        total = 0
        for unit in mo_units:
            total += self._strategy.calculate(
                unit, period, unique=True)
        return total
